# **********************************************************************
# Copyright 2018 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import fcntl
import fnmatch
import json
import netaddr
import nsx_logging
import os
import pprint
import re
try:
    import simplejson
except ImportError:
    import json as simplejson
import subprocess
from threading import Timer
import time
import uuid

from cli.commands.cli_command_provider import CliCommandProvider
from cli.commands.common.arguments import get_interface_names
import cli.commands.host_shared.ovs_util as ovs_util
from cli.utils.command_utils import invoke_command
from cli.utils.error_code import ErrorCode as ec
from cli.commands.common.get_managers import get_managers
from cli.commands.lcp_common.nsx_proxy.controller_info import NestDBCcpSessionConvertor
from cli.utils.node_utils import get_node_version
from cli.utils.location_utils import FilePaths

nsx_logging.basicConfig(filename="/var/log/troubleshoot.log", lockname="rotate.lock",
                        maxBytes=8*1024*1024, backupCount=4, compress="gz", fmt="%(message)s")
lg = nsx_logging.getLogger("cli.commands.host_kvm.troubleshoot")
_SUPPORT_URL = "http://www.vmware.com/support.html"
KVM_CORE_DUMP_DIR = "/var/log/core"


def get_ccp_session():
    cmd = ["/etc/init.d/nsx-proxy", "status"]
    rc, out, err = invoke_command(cmd);
    if rc != 0:
        return -1, "MSG_HOST_CONTROLLER_INFO_GET_FAILED"

    ccp_session = NestDBCcpSessionConvertor.get_ccp_session_info()
    if ccp_session is None:
        return -1, "MSG_HOST_CONTROLLER_INFO_GET_FAILED"
    ccp_session = json.loads(ccp_session)
    return 0, ccp_session


def get_manager_session():
    mgrs = get_managers(CliCommandProvider)
    if mgrs is None:
        return -1, "MSG_UNABLE_RETRIEVE_MANAGER_CONFIG"
    return 0, mgrs


def get_agent_status():
    status = {}
    agents = ['nsx-agent', 'nsx-exporter', 'nsx-mpa', 'nsx-nestdb',
              'nsx-opsagent', 'nsx-platform-client', 'nsx-proxy', 'nsx-sfhc']
    rv = 0
    for agent in agents:
        cmd = ["/etc/init.d/%s" % agent, "status"]
        rc, out, err = invoke_command(cmd)
        if rc != 0:
            rv = -1
            status[agent] = "INACTIVE"
        else:
            status[agent] = "ACTIVE"
    return rv, status


def run_ovs_vsctl(command, args=[], fmts=[], returnDict=True):
    ovs_vsctl_cmd = os.path.join(FilePaths.OVS_BIN_DIR, 'ovs-vsctl')
    if isinstance(fmts, list):
        cmd_list = [ovs_vsctl_cmd] + fmts + [command]
    else:
        cmd_list = [ovs_vsctl_cmd, str(fmts), command]
    if isinstance(args, list):
        cmd_list = cmd_list + args
    else:
        cmd_list = cmd_list + [str(args)]
    result_list = ovs_util.run_command(cmd_list, False, False)
    return result_list


def get_ovs_errors():
    # Use ip addr show
    arg='grep ERROR /var/log/openvswitch/*'
    p=subprocess.Popen(arg,shell=True,stdout=subprocess.PIPE)
    data = p.communicate()
    return data[0]


def get_ip_address(netdev='nsx-vtep0.0'):
    # Use ip addr show
    arg='ip addr show ' + netdev  + ' | grep \"inet \"'
    p=subprocess.Popen(arg,shell=True,stdout=subprocess.PIPE)
    data = p.communicate()
    if data[0] == '':
        return data[0]
    ipaddr = data[0].strip().split(' ')[1].split('/')[0]
    return ipaddr


def get_tunnel_status(errors):
    sessions={}
    interfaces = ovs_util.run_ovs_vsctl("list", "interface",
                   fmts=["--format=table", "--column=name"], returnDict=False)
    if interfaces == None:
        errors.append("\n%% Could not get tunnel status\n")
        return
    for interface in interfaces:
        if_name = interface[0].strip('"')
        if if_name.startswith('geneve'):
            bfd = ovs_util.run_ovs_vsctl("get", ["interface", if_name, "bfd"],
                                         returnDict=False)
            bfd_session={}
            for attr in bfd[0]:
                key = attr.strip('{')
                key = key.strip('}')
                key = key.strip(',')
                key = key.split('=')
                val = key[1].strip('"')
                key = key[0]
                bfd_session[key] = val
            bfd_status = run_ovs_vsctl("get", ["interface", if_name,
                                       "bfd_status"], returnDict=False)
            bfd_status = [x.strip() for x in bfd_status[0].split(',')]
            for attr in bfd_status:
                key = attr.strip('{')
                key = key.strip('}')
                key = key.strip(',')
                key = key.split('=')
                val = key[1].strip('"')
                key = key[0]
                bfd_session[key] = val
            if bfd == None or bfd_status == None:
                errors.append("%% No BFD session found on interface %s\n" %
                              if_name)
            if bfd != None and bfd_session['state'] != 'up':
                errors.append("%% BFD session on interface %s, %s<->%s is down" %
                              (if_name, bfd_session["bfd_src_ip"],
                                bfd_session["bfd_dst_ip"]))
            sessions[bfd_session["bfd_dst_ip"]] = bfd_session
    lg.info("\n%% Geneve Tunnel BFD Status:\n%s",
            json.dumps(sessions, indent = 4, sort_keys = True))


def get_core_dump_location():
    p = subprocess.Popen("cat /proc/sys/kernel/core_pattern",
                         stdout=subprocess.PIPE, shell=True)
    core_loc, _ = p.communicate()
    core_loc = core_loc.decode('utf-8')
    if core_loc[0] == "/":
        delm = core_loc.rfind("/")
        core_loc = str(core_loc[:delm+1])
    elif "apport" in core_loc:
        # If core_pattern is configured to pipe to apport,
        # apport saves the core dump in /var/crash
        core_loc = "/var/crash/"
    else:
        core_loc = None
    return core_loc


def find_crash_backtrace(errors, ncrash, filename):
    syslog_file = open(filename, 'r')
    for line in syslog_file:
        if 'Caught' in line:
            bt = line.strip('\n').replace('#012', '\n')
            ncrash[0] += 1
            errors.append("\n%% Crash: %d  Backtrace:\n%s" % (ncrash[0], bt))


class LcpKvmTroubleshooter(CliCommandProvider):
    def __call__(self, cmd_list):
        result = {}
        errors = []
        # manager connection
        rc, managers = get_manager_session()
        if rc == -1:
            errors.append("\n%% Not connected to nsx-manager\n")
        else:
            lg.info("\n%% Connected to nsx-manager:\n%s",
                    json.dumps(managers, indent = 4, sort_keys = True))

        # controller connection
        rc, controllers = get_ccp_session()
        if rc == -1:
            errors.append("\n%% Not connected to nsx-controller\n")
        else:
            lg.info("\n%% Connected to nsx-controller:\n%s",
                    json.dumps(controllers, indent = 4, sort_keys = True))

        # core dumps
        core_loc = get_core_dump_location()
        if core_loc:
            core_dumps = [fn for fn in os.listdir(core_loc) if 'core' in fn or 'crash' in fn]
            core_index = list(range(1, len(core_dumps)+1))
            errors.append("\n%% Found %d core files:" % len(core_dumps))
            errors.append("\n".join(' ' + str(a) + '. ' + core_loc + b
                          for a,b in zip(core_index, core_dumps)))
        else:
            lg.info("\n%% No core files found\n")

        # Backtrace of any crash seen - this may not corelate to core file order
        ncrash = [0]
        if os.path.isfile('/var/log/vmware/nsx-syslog'):
            find_crash_backtrace(errors, ncrash, '/var/log/vmware/nsx-syslog')
        else:
            nsxa_logs = os.listdir("/var/log/")
            for logfile in nsxa_logs:
                if (logfile.startswith('syslog') or
                    logfile.startswith('message')):
                    find_crash_backtrace(errors, ncrash, '/var/log/' + logfile)

        # check for all agents' state
        rc, agents = get_agent_status()
        if rc == -1:
            errors.append("\n%% Some nsx lcp agents are not active: %s\n" %
                          json.dumps(agents, indent = 4, sort_keys = True))
        else:
            lg.info("\n%% LCP Agents' status:\n%s",
                        json.dumps(agents, indent = 4, sort_keys = True))

        # get nsx-agent ovs version
        nsxa_ovs_ver = ovs_util.run_nsxa_ctl("version", header=False)
        if nsxa_ovs_ver == None:
            errors.append("\n%% nsx-agent is not active\n")
            # get ovs version
            ovs_ver = ovs_util.run_ovs_vsctl("--version", returnDict=False)
            if ovs_ver != None:
                lg.info("\n%% OVS version: %s\n", ovs_ver[0][3]);
        else:
            lg.info("\n%% libopenvswitch version: %s\n", nsxa_ovs_ver[0][2]);
            lg.info("\n%% OVS Version: %s\n", nsxa_ovs_ver[1][3]);

        # get nsx host version
        version = get_node_version()
        if not version:
            errors.append("\n%% nsx-lcp version unknown\n")
        else:
            lg.info("\n%% nsx-lcp version %s\n", version)

        # ovs bridge
        ovs_br = ovs_util.run_ovs_vsctl("list", "bridge",
                    fmts=["--format=table", "--column=name,_uuid,external-ids"],
                    returnDict=False)
        if ovs_br != None:
            br_found = False
            for br in ovs_br:
                if 'nsx-managed' in br[2]:
                    lg.info("\n%% nsx-managed ovs bridge is [%s : %s]\n",
                            br[0], br[1])
                    br_found = True
                    break
            if br_found == False:
                errors.append("\n%% nsx-managed ovs bridge is missing\n")
        else:
            errors.append("\n%% ovs not running\n")

        # Check ovs related errors
        ovs_errors = get_ovs_errors()
        if ovs_errors != '':
            errors.append("\n%% Found errors in openvswitch logs:\n%s" %
                          ovs_errors)

        # vtep interface
        vtep_ip = get_ip_address()
        if vtep_ip == '':
            errors.append("\n%% VTEP interface does not have an IP address\n")
        else:
            lg.info("\n%% VTEP interface IP address %s\n", vtep_ip)

        # Geneve tunnels
        get_tunnel_status(errors)

        # load
        kill = lambda process: process.kill()
        cmd = ["vmstat", "-S", "M", "-w", "1", "5"]
        rc, out, err = invoke_command(cmd)
        load = out
        if rc != 0:
            load = err
        lg.info("\n%% Current system load (mem in MB):\n%s", load)

        # disk space
        cmd = ["df"]
        rc, out, err = invoke_command(cmd)
        disk_usage = out
        if rc != 0:
            disk_usage = err
        lg.info("\n%% Current disk space (mem in Bytes):\n%s", disk_usage)

        # nsx-agent to ovs communication
        missing_flows = ovs_util.run_nsxa_ctl('flow/list', 'missing',
                                              split = False, header = False)
        if missing_flows == None:
            pass
        elif len(missing_flows):
            errors.append("\n%% OVS is missing %d flows\n" % len(missing_flows))

        extra_flows = ovs_util.run_nsxa_ctl('flow/list', 'extra',
                                              split = False, header = False)
        if extra_flows == None:
            pass
        elif len(extra_flows):
            errors.append("\n%% OVS has %d extra flows\n" % len(extra_flows))

        # DFW consistency
        # dfw = LcpKvmDfwConsistencyCheck(CliCommandProvider)
        # dfw(['verify', 'firewall', 'state'])
        result["results"] = errors
        return "OK\n\n%s" % simplejson.dumps(result)


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


class LcpKvmDfwConsistencyCheck(CliCommandProvider):
    def __init__(self, CliCommandProvider):
        self.dfw_addrsets = {}
        self.dfw_rules = {}
        self.dfw_vif_rules = {}
        self.dfw_vifs = {}


    def check_ip_address_consistency(self, rule_id, address_list,
                                     dfw_address_list):
        if 'ip_address' in address_list:
            for ip_address in address_list['ip_address']:
                if 'ipv4' in ip_address:
                    address = netaddr.IPAddress(ip_address['ipv4'])
                    hostlen = 32
                else:
                    address = netaddr.IPAddress(ip_address['ipv6'])
                    hostlen = 128

                address = str(str(address), "utf-8")
                if ip_address['prefix_length'] == hostlen:
                    if address not in dfw_address_list:
                        errors.append("%% Missing to ip address in rule %d" %
                                      rule_id)
                        return False
                else:
                    prefix = address + '/' + str(ip_address['prefix_length'])
                    if prefix not in dfw_address_list:
                        errors.append("%% Missing to ip prefix in rule %d" %
                                      rule_id)
                        return False
        if 'container' in address_list:
            for container in address_list['container']:
                container_uuid = str(str(uuid.UUID(
                                    int=(container["left"] << 64) |
                                         container["right"])), "utf-8")
                if container_uuid not in dfw_address_list:
                    errors.append("%% Missing to addrset %s in rule %d" %
                                  rule_id)
                    return False
        return True


    def dfw_rule_consistency_check(self, vif_id, nestdb_rule):
        # check each field in nestdb rule is consistent with rule in dfw
        rule_id = nestdb_rule['id']
        if rule_id not in self.dfw_rules:
            return False
        dfw_rule = self.dfw_rules[rule_id]

        if vif_id not in dfw_rule['applied_to']:
            errors.append("\n%% Rule id %d not applied to vif %s" %
                          (rule_id, vif_id))
            return False

        exp_action = str(nestdb_rule['action'])
        action = str(dfw_rule['action'])
        if ((exp_action == 'RA_DROP' and action != 'drop') or
            (exp_action == 'RA_PASS' and action != 'accept') or
            (exp_action == 'RA_REJECT' and action != 'reject') or
            (exp_action == 'RA_REDIRECT' and action != 'redirect')):
            errors.append("\n%% Rule id %d, inconsistent action, vif %s (%s != %s)" %
                          (rule_id, vif_id, exp_action, action))
            return False

        exp_dir = str(nestdb_rule['direction'])
        direction = str(dfw_rule['direction'])
        if ((exp_dir == 'RD_BOTH' and direction != 'inout') or
            (exp_dir == 'RD_IN' and direction != 'in') or
            (exp_dir == 'RD_OUT' and direction != 'out')):
            errors.append("\n%% Rule id %d, inconsistent direction, vif %s" %
                          (rule_id, vif_id))
            return False
        dfw_rule['applied_to'].remove(vif_id)

        if 'from_address' in nestdb_rule:
            if self.check_ip_address_consistency(rule_id,
                nestdb_rule['from_address'], dfw_rule['from']) == False:
                return False
            if nestdb_rule['from_address']['negate'] != dfw_rule['from_negate']:
                errors.append("\n%% Rule id %d, inconsistent from address" %
                              rule_id)
                return False

        if 'to_address' in nestdb_rule:
            if self.check_ip_address_consistency(rule_id,
                nestdb_rule['to_address'], dfw_rule['to']) == False:
                return False
            if nestdb_rule['to_address']['negate'] != dfw_rule['to_negate']:
                errors.append("\n%% Rule id %d, inconsistent to address" %
                              rule_id)
                return False

        return True


    def __call__(self, cmd_list):
        result = {}
        errors = []
        # nestdb rules
        cmd = ['/opt/vmware/nsx-nestdb/bin/nestdb-cli', '--json', '--cmd',
                'get vmware.nsx.nestdb.RuleMsg']
        rc,out,err = invoke_command(cmd)
        nestdb_rules=json.loads(out)["result"]

        # dfw addrsests
        out = ovs_util.run_nsxa_ctl('dfw/addrsets', '-json', split = False,
                                         header = False)
        vif_addrsets = json.loads(out[0])['dfw_vif_addrsets']
        for vif_addrset in vif_addrsets:
            addrsets = vif_addrset['addrsets']
            for addrset in addrsets:
                uuid = addrset['uuid']
                if uuid not in self.dfw_addrsets:
                    self.dfw_addrsets[uuid] = addrset['addresses']

        # dfw vifs
        out = ovs_util.run_nsxa_ctl('dfw/vif', '-json', split = False,
                                    header = False)
        self.dfw_vifs = json.loads(out[0]).get('dfw_vif_list')[1:]

        # dfw vif rules
        for vif in self.dfw_vifs:
            vif_id = vif['vif_id']
            args = [vif_id, '-json']
            out = ovs_util.run_nsxa_ctl('dfw/rules', args,
                                        split = False, header = False)
            self.dfw_vif_rules[vif_id] = json.loads(out[0]).get(
                                            'dfw_vif_rules')[0]['rulesets']

        # coalesce same rules across all vifs
        for vif in self.dfw_vifs:
            vif_id = vif['vif_id']
            vif_ruleset = self.dfw_vif_rules[vif_id]
            for vif_rules in vif_ruleset:
                for nsxa_rule in vif_rules['rules']:
                    nsxa_rule = nsxa_rule.rstrip(';').split()
                    rule_id = int(nsxa_rule[1])
                    if rule_id not in self.dfw_rules:
                        self.dfw_rules[rule_id] = {}

                    rule = self.dfw_rules[rule_id]

                    if 'applied_to' not in rule:
                        rule['applied_to'] = set([vif_id])
                    else:
                        rule['applied_to'].add(vif_id)

                    direction = nsxa_rule[2]
                    if 'direction' in rule and rule['direction'] != direction:
                        errors.append('\n%% Inconsistent direction in rule %d' %
                                      rule_id)
                        return 'FAILURE'
                    else:
                        rule['direction'] = direction

                    proto = nsxa_rule[4]
                    if 'protocol' in rule and rule['protocol'] != proto:
                        errors.append('\n%% Inconsistent protocol in rule %d' %
                                      rule_id)
                        return 'FAILURE'
                    else:
                        rule['protocol'] = proto

                    action = ''
                    if 'accept' in nsxa_rule:
                        action = 'accept'
                    elif 'drop' in nsxa_rule:
                        action = 'drop'
                    elif 'reject' in nsxa_rule:
                        action = 'reject'
                    elif 'redirect' in nsxa_rule:
                        action = 'redirect'

                    if 'action' in rule and rule['action'] != action:
                        errors.append('\n%% Inconsistent action in rule %d' %
                                      rule_id)
                        return 'FAILURE'
                    else:
                        rule['action'] = action

                    from_idx = nsxa_rule.index("from")
                    if 'from' not in rule:
                        rule['from'] = set()
                    if nsxa_rule[from_idx - 1] == 'not':
                        if ('from_negate' in rule and
                            rule['from_negate'] == False):
                            errors.append('\n%% Inconsistent from addrs in rule %d' % rule_id)
                            return FAILURE
                        else:
                            rule['from_negate'] = True
                    else:
                        if ('from_negate' in rule and
                            rule['from_negate'] == True):
                            errors.append('\n%% Inconsistent from addrs in rule %d' % rule_id)
                            return FAILURE
                        else:
                            rule['from_negate'] = False

                    if nsxa_rule[from_idx + 1] == 'addrset':
                        addrset_uuid = nsxa_rule[from_idx + 2]
                        rule['from'].add(addrset_uuid)
                        """
                        if addrset_uuid not in self.dfw_addrsets:
                            lg.info('\n%% Inconsistent from addrs in rule %d', rule_id)
                            return FAILURE
                        rule['from'].update(self.dfw_addrsets[addrset_uuid])
                        """
                    elif nsxa_rule[from_idx + 1] == 'any':
                        if len(rule['from']) == 0:
                            rule['from'].add('any')
                        elif len(rule['from']) > 1 or 'any' not in rule['from']:
                            errors.append('\n%% Inconsistent from addrs in rule %d' % rule_id)
                            return FAILURE
                    else:
                        rule['from'].add(nsxa_rule[from_idx + 1])

                    to_idx = nsxa_rule.index("to")
                    if 'to' not in rule:
                        rule['to'] = set()

                    if nsxa_rule[to_idx - 1] == 'not':
                        if 'to_negate' in rule and rule['to_negate'] == False:
                            errors.append('\n%% Inconsistent to addrs in rule %d' % rule_id)
                            return FAILURE
                        else:
                            rule['to_negate'] = True
                    else:
                        if 'to_negate' in rule and rule['to_negate'] == True:
                            errors.append('\n%% Inconsistent TO addrs in rule %d' % rule_id)
                            return FAILURE
                        else:
                            rule['to_negate'] = False

                    if nsxa_rule[to_idx + 1] == 'addrset':
                        addrset_uuid = nsxa_rule[to_idx + 2]
                        rule['to'].add(addrset_uuid)
                        """
                        if addrset_uuid not in self.dfw_addrsets:
                            print '\n%% Inconsistent To addrs in rule %d' %(
                                rule_id)
                            return FAILURE
                        rule['to'].update(self.dfw_addrsets[addrset_uuid])
                        """
                    elif nsxa_rule[to_idx + 1] == 'any':
                        if len(rule['to']) == 0:
                            rule['to'].add('any')
                        elif len(rule['to']) > 1 or 'any' not in rule['to']:
                            errors.append('\n%% Inconsistent to addrs in rule %d' % rule_id)
                            return FAILURE
                    else:
                        rule['to'].add(nsxa_rule[to_idx + 1])

        lg.info("%% DFW state: \n%s" %
                json.dumps(self.dfw_rules, indent = 4, sort_keys = True, cls=SetEncoder))

        # do a consistency check between nestdb and dfw states
        rv = True
        for entry in nestdb_rules:
            rule = entry['value']
            if 'applied_to' in rule:
                for vif_id in rule['applied_to']['vif']:
                    if vif_id not in self.dfw_vif_rules:
                        continue
                    if self.dfw_rule_consistency_check(vif_id, rule) == False:
                        rv = False

            else:
                for vif in self.dfw_vifs:
                    vif_id = vif['vif_id']
                    if self.dfw_rule_consistency_check(vif_id, rule) == False:
                        rv = False

        # Check if any extra config present
        for rule_id, rule in list(self.dfw_rules.items()):
            if len(rule['applied_to']) == 0:
                continue
            errors.append("\n%% Found suprious rule %d for vifs:\n" % rule_id)
            for vif_id in rule['applied_to']:
                errors.append(" %s" % vif_id)

        if rv:
            errors.append("\n%% Distributed Firewall state is consistent\n")
        else:
            errors.append("\n%% Distributed Firewall state is inconsistent\n")
        result["results"] = errors
        return "OK\n\n%s" % simplejson.dumps(result)
