# ****************************************************************************
# Copyright 2017, 2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ****************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ****************************************************************************
import nsx_logging
import signal
import sys
import time
import uuid

from cli.utils import rest_utils
from cli.utils.command_utils import get_all_modifiers
from cli.utils.constants import NODE_TYPE_GLOBAL_MANAGER
from cli.utils.constants import NODE_TYPE_MANAGER
from cli.utils.constants import RESOURCE_TYPE_CLUSTER_NODE
from cli.utils.error_code import ErrorCode as ec
from cli.utils.node_utils import get_node_type
from compat23 import compat23_httplib


lg = nsx_logging.getLogger(__name__)
# The Central API feature is only supported on Manager and GM nodes and
# depends on modules only deployed on these node types
_NODE_TYPE = get_node_type()
if NODE_TYPE_MANAGER in _NODE_TYPE or NODE_TYPE_GLOBAL_MANAGER in _NODE_TYPE:
    try:
        _NSX_RPC_PROTOPATH = "/opt/vmware/nsx-monitoring/python"
        if _NSX_RPC_PROTOPATH not in sys.path:
            sys.path.append(_NSX_RPC_PROTOPATH)
        from vmware.nsx.messaging.generic_unified_messaging_pb2 import      \
            GenericUnifiedMsg
        from vmware.nsx.platform_management import common_pb2
        from vmware.nsx.platform_management.platform_client_services_pb2 import \
            PlatformMgmtHostService_Stub, PlatformMgmtMpUnifiedMsg,             \
            PlatformMgmtHostUnifiedMsg
        from vmware.nsx.rpc import NsxRpcClient
        from vmware.nsx.rpc import NsxRpcConnection
        from vmware.nsx.rpc.client.channel import NsxRpcCallError
    except Exception:
        lg.exeception(
            "Error importing modules required by the central cli feature",
            error_code=ec.INTERNAL_ERROR)
    _STATUS_VALUE_TO_STRING = {
        common_pb2.STATUS_INVALID: "STATUS_INVALID",
        common_pb2.STATUS_RUNNING: "STATUS_RUNNING",
        common_pb2.STATUS_COMPLETED: "STATUS_COMPLETED",
        common_pb2.STATUS_SCHEDULED: "STATUS_SCHEDULED",
        common_pb2.STATUS_CANCELING: "STATUS_CANCELING",
        common_pb2.STATUS_CANCELED: "STATUS_CANCELED",
        common_pb2.STATUS_NOT_FOUND: "STATUS_NOT_FOUND",
        common_pb2.STATUS_CONFLICT: "STATUS_CONFLICT"}

ASYNC_POOL_SIZE = 5
DELAY_BETWEEN_QUERY_REQUESTS = 2  # Number of seconds between status queries

_CENTRAL_CLI_ARGUMENT_CHECK_REQUEST_TYPE = (
    "com.vmware.nsx.platform_management.CentralCliArgumentCheckRequestMsg")
_CENTRAL_CLI_CANCEL_REQUEST_TYPE = (
    "com.vmware.nsx.platform_management.CentralCliCancelRequestMsg")
_CENTRAL_CLI_COMMAND_REQUEST_TYPE = (
    "com.vmware.nsx.platform_management.CentralCliCommandRequestMsg")
_CENTRAL_CLI_STATUS_REQUEST_TYPE = (
    "com.vmware.nsx.platform_management.CentralCliStatusRequestMsg")
_CENTRAL_API_DEFAULT_TIMEOUT = 60
_CENTRAL_CLI_METHOD_NAME = "RpcMsgHandler"
_CLUSTER_NODE_CACHE = {}  # key: cluster_node_uuid, val: endpoint_id
_ENDPOINT_ID_FOUND = 0
_ENDPOINT_ID_NOT_FOUND = 1
_ENDPOINT_ID_SERVICE_UNAVAILABLE = 2
_LOCAL_CLUSTER_NODES_URI = "/cluster/nodes/%s"
_LOCAL_NSX_RPC_ENDPOINT = "unix:///var/run/vmware/appl-proxy/aph.sock"
_NSX_RPC_CLIENT_PREFIX = "nsx_rpc_client"
_NSX_RPC_ERROR_TIMEOUT = "TIMEOUT"


class CentralCLIAPIException(Exception):
    pass


def invoke_command(node, node_type, cmd, modifier, context,
                   username, rbac_permissions, debug_mode):
    '''Invoke a remote command

    Return the unique request id for command queued for execution.
    '''
    # Import pwd here instead of globally as pwd may not be available on all
    # platforms and is only needed on MP for processing Central CLI requests
    import pwd
    request_id = str(uuid.uuid4())  # Random and unique request id
    result, endpoint_id = _get_destination_node_endpoint_id(
        node, node_type, "central cli query")
    if result != _ENDPOINT_ID_FOUND:
        raise CentralCLIAPIException(
            "Error invoking central cli command, Cannot determine "
            "endpoint_id for node %s: %s", node, endpoint_id)
    host_msg, generic_unified_msg = _instantiate_request_message_instances()
    host_msg.object_type = _CENTRAL_CLI_COMMAND_REQUEST_TYPE
    host_msg.msg1.request_id = request_id
    host_msg.msg1.command = cmd
    if (modifier and modifier.split()[0] in get_all_modifiers()):
        host_msg.msg1.modifier = modifier
    host_msg.msg1.username = str(pwd.getpwnam(username).pw_uid)
    host_msg.msg1.context = context
    host_msg.msg1.debug_mode = debug_mode
    host_msg.msg1.rbac_permission.extend(rbac_permissions)
    generic_unified_msg.msg = host_msg.SerializeToString()
    result = _invoke_rpc(
        endpoint_id, _CENTRAL_CLI_METHOD_NAME, generic_unified_msg,
        timeout=_CENTRAL_API_DEFAULT_TIMEOUT)
    if result is None:
        err_msg = (
            "Error invoking central cli command on node: %s, request_id: "
            "%s" % (node, request_id))
        lg.error(err_msg, error_code=ec.INTERNAL_ERROR)
        raise CentralCLIAPIException(err_msg)
    return request_id


def query_request(node, node_type, request_id):
    '''Query status of remote command invocation

    Return a tuple of request status and output from request.
    '''
    result, endpoint_id = _get_destination_node_endpoint_id(
        node, node_type, "central cli query")
    if result != _ENDPOINT_ID_FOUND:
        raise CentralCLIAPIException(
            "Error querying central cli request, Cannot determine "
            "endpoint_id for node %s: %s", node, endpoint_id)
    host_msg, generic_unified_msg = _instantiate_request_message_instances()
    host_msg.object_type = _CENTRAL_CLI_STATUS_REQUEST_TYPE
    host_msg.msg2.request_id = request_id
    generic_unified_msg.msg = host_msg.SerializeToString()
    result = _invoke_rpc(
        endpoint_id, _CENTRAL_CLI_METHOD_NAME, generic_unified_msg,
        timeout=_CENTRAL_API_DEFAULT_TIMEOUT)
    if result is None:
        err_msg = (
            "Error querying central cli request on node: %s, request_id: "
            "%s" % (node, request_id))
        lg.error(err_msg, error_code=ec.INTERNAL_ERROR)
        raise CentralCLIAPIException(err_msg)
    response_msg = _instantiate_response_message_instance()
    response_msg.ParseFromString(result.msg)
    status = response_msg.msg1.status
    status_values = _get_status_value_dict()
    if status not in status_values:
        status = common_pb2.STATUS_INVALID
    output = response_msg.msg1.output.decode("utf8")
    return status_values[status], output


def _invoke_and_wait(node, node_type, cmd, modifier, context, username,
                     rbac_permissions, debug_mode):
    '''Invoke a cmd and block until it completes'''

    start_time = None
    # All cmds with 'get' should display datetime stamp in it's output
    if cmd.startswith("get"):
        start_time = time.time()

    request_id = invoke_command(node, node_type, cmd, modifier, context,
                                username, rbac_permissions, debug_mode)

    status, output = query_request(node, node_type, request_id)

    while (status in ('STATUS_RUNNING',
                      'STATUS_SCHEDULED',
                      'STATUS_CANCELING')):
        time.sleep(DELAY_BETWEEN_QUERY_REQUESTS)
        status, output = query_request(node, node_type, request_id)

    return status, node, output, modifier, start_time


def worker_init():
    # Ignore SIGINT in worker process
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def async_invoke_and_wait(nodes, cmd, modifier, context, username,
                          rbac_permissions, debug_mode, callback):
    '''Invoke command on nodes asynchrousnously and run callback on finish'''
    # Importing multiprocessing lazily, as ESX doesn't have this module as part
    # of its Python installation. This is never a problem as only managers need
    # multiprocessing module.
    from multiprocessing import Pool
    pool = Pool(ASYNC_POOL_SIZE, worker_init)

    results = [pool.apply_async(_invoke_and_wait,
                                args=(node, node_type, cmd, modifier, context,
                                      username, rbac_permissions, debug_mode),
                                callback=callback) for node, node_type in nodes]

    try:
        # If timeout is not specified, interrupt doesn't get delivered. This is
        # due to a Python bug http://bugs.python.org/issue8296. By specifying a
        # high timeout, we ensure that interrupt is received.
        [res.get(999999) for res in results]
    finally: #  Ensure that all pool workers exit
        # TODO Also request cancellation on remote node using cancel_request, if
        # the API call(s) didn't finish.
        pool.terminate()
        pool.join()


def _check_argument(node, node_type, action, context, arg_id, user_input, value,
                    username, debug_mode):
    import pwd
    result, endpoint_id = _get_destination_node_endpoint_id(
        node, node_type, "central cli check argument")
    if result != _ENDPOINT_ID_FOUND:
        lg.error(
            "Error checking central cli argument, Cannot determine "
            "endpoint_id for node %s: %s", node, endpoint_id,
            error_code=ec.INTERNAL_ERROR)
        return False, None
    host_msg, generic_unified_msg = _instantiate_request_message_instances()
    host_msg.object_type = _CENTRAL_CLI_ARGUMENT_CHECK_REQUEST_TYPE
    host_msg.msg4.action = action
    host_msg.msg4.arg_id = arg_id
    host_msg.msg4.context = context
    host_msg.msg4.debug_mode = debug_mode
    host_msg.msg4.user_input = user_input
    host_msg.msg4.value = value
    host_msg.msg4.username = str(pwd.getpwnam(username).pw_uid)
    generic_unified_msg.msg = host_msg.SerializeToString()
    result = _invoke_rpc(
        endpoint_id, _CENTRAL_CLI_METHOD_NAME, generic_unified_msg,
        timeout=_CENTRAL_API_DEFAULT_TIMEOUT)
    if result is None:
        lg.error(
            "Error checking central cli argument on node: %s, arg_id: "
            "%s" % (node, arg_id), error_code=ec.INTERNAL_ERROR)
        return False, None
    response_msg = _instantiate_response_message_instance()
    response_msg.ParseFromString(result.msg)
    result = response_msg.msg3.result
    arguments = []
    pos = 0
    size = len(response_msg.msg3.description)
    for argument in response_msg.msg3.value:
        desc = response_msg.msg3.description[pos] if pos < size else ""
        arguments.append({
            "value": argument,
            "description": desc})
        pos += 1
    return result, arguments


def get_arg_values(node, node_type, context, arg_id, user_input, value,
                   username, debug_mode):
    '''Get valid values for a enum or query arg'''
    result, arguments = _check_argument(node, node_type, 'get_arg_values',
                                        context, arg_id, user_input, value,
                                        username, debug_mode)
    if not result:
        raise CentralCLIAPIException('get_arg_values failed')
    else:
        return arguments


def verify_arg(node, node_type, context, arg_id, user_input, value, username,
               debug_mode):
    '''Verify a verify_arg'''
    result, _ = _check_argument(node, node_type, 'verify_arg', context,
                                arg_id, user_input, value,
                                username, debug_mode)
    if not result:
        raise CentralCLIAPIException('verify_arg_values failed')
    return result


def cancel_request(node, node_type, request_id):
    '''Cancel a remote command'''
    result, endpoint_id = _get_destination_node_endpoint_id(
        node, node_type, "central cli cancel")
    if result != _ENDPOINT_ID_FOUND:
        raise CentralCLIAPIException(
            "Error cancelling central cli request, Cannot determine "
            "endpoint_id for node %s: %s", node, endpoint_id)
    host_msg, generic_unified_msg = _instantiate_request_message_instances()
    host_msg.object_type = _CENTRAL_CLI_CANCEL_REQUEST_TYPE
    host_msg.msg3.request_id = request_id
    generic_unified_msg.msg = host_msg.SerializeToString()
    result = _invoke_rpc(
        endpoint_id, _CENTRAL_CLI_METHOD_NAME, generic_unified_msg,
        timeout=_CENTRAL_API_DEFAULT_TIMEOUT)
    if result is None:
        err_msg = (
            "Error cancelling central cli request on node: %s, request_id: "
            "%s" % (node, request_id))
        lg.error(err_msg, error_code=ec.INTERNAL_ERROR)
        raise CentralCLIAPIException(err_msg)


def _get_destination_node_endpoint_id(node, node_resource_type, req_label):
    """Get endpoint_id of specified node, returns
       (_ENDPOINT_ID_FOUND, endpoint_id) tuple if successful; otherwise,
       (error_code, None) tuple where error_code can be:
       - _ENDPOINT_ID_NOT_FOUND
       - _ENDPOINT_ID_SERVICE_UNAVAILABLE"""
    # The endpoint_id of non-Manager/GM nodes is always the same as node id
    if node_resource_type != RESOURCE_TYPE_CLUSTER_NODE:
        return _ENDPOINT_ID_FOUND, node
    # Use cached endpoint_id value if previously fetched, per the messaging
    # team a registered cluster node's endpoint_id never changes
    if node in _CLUSTER_NODE_CACHE:
        return _ENDPOINT_ID_FOUND, _CLUSTER_NODE_CACHE[node]
    # Otherwise, to find the endpoint_id of a Manager/GM node, query
    # policy/proton for the specified node's endpoint_id, note: return
    # _ENDPOINT_ID_SERVICE_UNAVAILABLE to caller if policy/proton
    # unavailable so it can differentiate between invalid node versus retry.
    uri = _LOCAL_CLUSTER_NODES_URI % node
    status, obj, text, resp = rest_utils.local_node_request(
        "GET", uri, proton_request=True)
    if status == compat23_httplib.OK:
        endpoint_id = obj.get(
            "manager_role", {}).get("appliance_connection_info", {}).get(
                "service_endpoint_uuid")
        if endpoint_id is not None:
            _CLUSTER_NODE_CACHE[node] = endpoint_id
            return _ENDPOINT_ID_FOUND, endpoint_id
    lg.info(
        "Unable to process %s request because the manager node's endpoint_id "
        "can't be determined, status: %s, obj: %s", req_label, status, str(obj))
    if status == compat23_httplib.SERVICE_UNAVAILABLE:
        return _ENDPOINT_ID_SERVICE_UNAVAILABLE, None
    return _ENDPOINT_ID_NOT_FOUND, None


def _invoke_rpc(destination_endpoint_id, method_name, arg, timeout=60):
    """Invoke RPC (stub/method_name) on remote node identified by
       destination_endpoint_id, return response message if successful;
       otherwise, None.
    """
    try:
        conn = NsxRpcConnection()
        conn.Connect(_LOCAL_NSX_RPC_ENDPOINT)
        with NsxRpcClient(
                PlatformMgmtHostService_Stub, connection=conn, timeout=timeout,
                destination=destination_endpoint_id) as nsx_rpc_client:
            method = eval("%s.%s" % (_NSX_RPC_CLIENT_PREFIX, method_name))
            return method(arg)
    except ConnectionRefusedError as ex:
        lg.error(
            "APH connection refused when invoking %s on remote node %s",
            method_name, destination_endpoint_id, error_code=ec.INTERNAL_ERROR)
    except NsxRpcCallError as ex:
        if ex.name == _NSX_RPC_ERROR_TIMEOUT:
            lg.error(
                "%s invocation on remote node %s timed out",
                method_name, destination_endpoint_id,
                error_code=ec.INTERNAL_ERROR)
        else:
            lg.error(
                "Error invoking method %s on remote node %s code %s status: %s",
                method_name, destination_endpoint_id, ex.code, ex.status,
                error_code=ec.INTERNAL_ERROR)
    except Exception:
        lg.exception(
            "Error invoking method %s on remote node %s",
            method_name, destination_endpoint_id, error_code=ec.INTERNAL_ERROR)
    finally:
        conn.Close()
    return None


def _instantiate_request_message_instances():
    """Return new PlatformMgmtHostUnifiedMsg and GenericUnifiedMsg instances.

       Note: this method allows unit tests to mock protobuf message instances.
    """
    return PlatformMgmtHostUnifiedMsg(), GenericUnifiedMsg()


def _instantiate_response_message_instance():
    """Return a new PlatformMgmtMpUnifiedMsg instance

       Note: this method allows unit tests to mock protobuf message instances.
    """
    return PlatformMgmtMpUnifiedMsg()


def _get_status_value_dict():
    return _STATUS_VALUE_TO_STRING
