#!/bin/sh
#
# Watchdog script that manages VMware services.
# Launches the specified process, and respawns it after it exits;
# Gives up after recording the specified number of 'quick failures'
# in succession or after recording a specified total number of
# failures (over any length of time).
#
# This script supports Linux and VMkernel platforms and
# should be tested on both.

# Handler used to cleanly exit the watchdog
cleanup () {
   log daemon.info "[$$] Signal received: exiting the watchdog"
   if [ $START = 1 ] ; then
       rm -rf ${PIDFILE}
   fi
   exit 0
}

# Trap all trappable signals (this excludes 9 17 19 23) and clean up.
trap cleanup 1 2 3 6 7 8 10 13 14 15 16 24 25 26 27 30 31

usage () {
    echo "Usage: $0 [-n] -s|-k|-r <tag> [(options)] <command>"
    echo "   Start the watchdog: $0 -s <tag> [-u <min_uptime>] [-q <max_quick_failures>] [-t <max_total_failures>] [-p <custom_pid_file>] <command>"
    echo "   Start the watchdog in background: $0 -d -s <tag> [-u <min_uptime>] [-q <max_quick_failures>] [-t <max_total_failures>] [-p <custom_pid_file>] <command>"
    echo "   Kill a running watchdog: $0 -k <tag>"
    echo "   Query whether a watchdog is running: $0 -r <tag>"
    echo "   Suppress logging: -n"
    exit 1
}

if [ $# -lt 2 ]
then
    usage
fi

log () {
    if [ "${QUIET}" -ne 0 ]
    then
        return
    fi

    if [ $# -lt 2 ]
    then
        logger -p daemon.err -t NSX "Bad log usage."
    else
        _loglevel=$1
        _logmsg=$2
        shift; shift;
        _logoptarg=$@
        logger -p $_loglevel -t NSX "$_logmsg" $_logoptarg
    fi
}

start () {
    MSG="[$$] Begin '$CMD', min-uptime = $MIN_UPTIME, max-quick-failures = "
    MSG="$MSG$MAX_QUICK_FAILURES, max-total-failures = $MAX_TOTAL_FAILURES,"
    MSG="$MSG bg_pid_file = '$BG_PID_FILE'"
    echo $MSG
    log daemon.info "$MSG"
    echo $$ > ${PIDFILE}

    ORGCMD=$CMD

    while [ ! $QUICK_FAILURES -gt $MAX_QUICK_FAILURES -a ! $TOTAL_FAILURES -gt $MAX_TOTAL_FAILURES ]
      do
      # Erase the last BG_PID_FILE if it exists
      if [ "$BG_PROC" -eq 1 ] ; then
         rm -f "$BG_PID_FILE" > /dev/null 2>&1
      fi

      # Adjust group parameter if set (only ESX has configRP script)
      if [ -f /usr/lib/vmware/rp/bin/configRP -a "$(echo $ORGCMD | grep group=)" != "" ]; then
         GROUP=$(echo $ORGCMD | sed 's|.*\+\+[^ ]*group=\([a-zA-Z0-9\/\.\_]*\).*|\1|')

         ADJGROUP=$(/usr/lib/vmware/rp/bin/configRP getRP $GROUP)

         if [ "$ADJGROUP" == "" ]; then
            log daemon.info "Adjusted group for $GROUP is empty"
            return
         fi

         CMD=$(echo $ORGCMD | sed "s|group=$GROUP|group=$ADJGROUP|")
      fi

      log daemon.info "Executing '$CMD'"
      LAST=$(date +%s)
      rm -rf $CUSTOM_PID_FILE
      # Lauches the command in a different process group.  Helps with signal handling.
      setsid $CMD &

      # Wait for process to stop
      if [ "$BG_PROC" -eq 1 ] ; then
         local WAIT_TIME=0
         while [ ! -e "$BG_PID_FILE" -o $WAIT_TIME -gt $MAX_BG_PID_FILE_WAIT ]; do
            sleep 1
            WAIT_TIME=$(($WAIT_TIME + 1))
         done

         # check whether the process exist
         while kill -0 $(cat "$BG_PID_FILE") 2>/dev/null ; do sleep 1; done
      else
         if [ -n "$CUSTOM_PID_FILE" ] ; then
            # wait up to 15 seconds for pid file to be populated
            for i in `seq 0 15`; do
               sleep 1
              if [ -f $CUSTOM_PID_FILE ] ; then
                break
              fi
            done
            pid=`cat $CUSTOM_PID_FILE`
            while [ -e /proc/$pid ]; do
               sleep 1;
            done
         else
            pid=$!
            wait $pid
         fi
         rc=$?
      fi

      if [ ! -f "$RESTART_FILE" -a ! -f "$ALWAYS_RESTART_FILE" ] ; then
          TOTAL_FAILURES=$(expr $TOTAL_FAILURES + 1)
          NOW=$(date +%s)
          UPTIME=$(expr $NOW - $LAST)
          if [ $UPTIME -lt $MIN_UPTIME ] ; then
              QUICK_FAILURES=$(expr $QUICK_FAILURES + 1)
              LOG_MESSAGE="'$CMD' exited after $UPTIME seconds (quick failure $QUICK_FAILURES) $rc"
              log daemon.err "$LOG_MESSAGE"
          else
              QUICK_FAILURES=0
              log daemon.err "'$CMD' exited after $UPTIME seconds $rc"
          fi
          # Debug Log for PR1899936
          if [ $rc -eq 127 ] ; then
              result=$(echo $CMD | grep netopa)
              if [ "$result" != "" ] ; then
                  output="$(ldd /usr/bin/netopa)"
                  log daemon.err "${output}"
              fi
          fi
          if [ "$CLEANUP_CMD" != "" ] ; then
              log daemon.info "Executing cleanup command '$CLEANUP_CMD'"
              setsid $CLEANUP_CMD > /dev/null 2>&1
          fi
      else
          log daemon.info "Restart file detected, removing it."
          rm -rf "$RESTART_FILE" > /dev/null 2>&1
          if [ $? != 0 ] ; then
               log daemon.warning "unable to delete file $RESTART_FILE"
          fi
      fi
    done

    log daemon.err "End '$ORGCMD', failure limit reached"
    rm -rf "${PIDFILE}"

    if [ "$EXIT_CLEANUP_CMD" != "" ] ; then
       log daemon.info "Executing cleanup command '$EXIT_CLEANUP_CMD' before exiting."
       $EXIT_CLEANUP_CMD > /dev/null 2>&1
    fi
    exit 0
}


daemonize () {
   # start the watchdog in background
   rm -rf ${PIDFILE}
   # This files redirection is fix for PR#756931
   setsid $0 ${ORIG_OPTIONS} $CMD &
   local TIMEOUT=${MAX_DAEMONIZE_TIMEOUT}

   while [ ! -e ${PIDFILE} -a ${TIMEOUT} -gt 0 ] ; do
      sleep 1
      TIMEOUT=$(expr ${TIMEOUT} - 1)
   done

   if [ ! -e ${PIDFILE} ] ; then
      local MSG="Unable to verify ${TAG} started after ${MAX_DAEMONIZE_TIMEOUT} seconds"
      log daemon.err "${MSG}"
      echo "${MSG}"
   fi
}


# The awk selects the lines which lists watchdog being run with the
# specified TAG and is not the process itself and then extracts the
# PID.
LookupWatchdogPID () {
   local ps_args=""

   if [ "$(uname)" = "VMkernel" ] ; then
      ps_args="-cu";
   else
      # ww for wide output
      ps_args="axww";
   fi
   # Ignore the PID of this instance and the awk process should filter itself by *not* maching its parameters.
   # The regex expects a space before "-s", but there is a "+" there in the awk parameter that is outputed by ps.
   # Be very careful when modifying this line not to make a regex that matches itself!
   ps $ps_args 2> /dev/null | awk "!/^ *$$ / && /\/$(basename $0)\>.* +-s $TAG\>/ { print \$1 }"
}

VerifyWatchdogPID () {
  [ "${1}" = "$(LookupWatchdogPID)" ]
}

GetPid () {
    if [ ! -f ${PIDFILE} ]; then
        log daemon.info "PID file $PIDFILE does not exist" -s
        echo 0
        return
    fi
    PID=$(cat ${PIDFILE} 2> /dev/null)
    if [ -z "${PID}" ] ; then
        # This is a measure of last resort in case the pid file has been deleted somehow
        # Not foolproof but should work most of the time
        PID=$(LookupWatchdogPID)
        if [ -n "${PID}" ] ; then
            log daemon.info "Found running watchdog with pid $PID" -s
        else
           log daemon.info "Neither $PIDFILE nor running watchdog found" -s
        fi
    else
        # PR 253106
        if VerifyWatchdogPID ${PID}  ; then
            PID=$(LookupWatchdogPID)
            log daemon.info "Watchdog for $TAG is now $PID"
        fi
    fi
    # We are sure that the PID is not us
    if [ -n "${PID}" ] ; then
        echo $PID
    else
        echo 0
    fi
}

stop () {
    PID=$(GetPid)
    if [ "${PID}" -eq 0 ]; then
        MSG="Unable to terminate watchdog: No running watchdog process for $TAG"
        log daemon.info "$MSG" -s
        exit 1
    else
       local TIMEOUT=${MAX_SHUTDOWN_TIMEOUT}
       MSG="Terminating watchdog process with PID $PID"
       log daemon.info "$MSG" -s
       kill -HUP $PID 2> /dev/null
       while [ $TIMEOUT -gt 0 ] && kill -9 $PID 2>/dev/null ; do
           TIMEOUT=$((TIMEOUT-1))
           sleep 1
       done
       return $((TIMEOUT == 0))
    fi
}

query () {
    PID=$(LookupWatchdogPID)
    echo "${PID}"
    if [ "${PID}" = "" ]; then
       return 1
    else
       return 0
    fi
}

# Give these variables default values.
# These can be overridden with command line arguments.
# 1,000,000 here is basically meant to be infinity.
MAX_QUICK_FAILURES=1
MAX_TOTAL_FAILURES=1000000
MIN_UPTIME=60
MAX_BG_PID_FILE_WAIT=10       # Wait 10 seconds for BG PID FILE to show up
MAX_DAEMONIZE_TIMEOUT=10
MAX_SHUTDOWN_TIMEOUT=5

# If this is specified, the pid to monitor is stored in this file.
CUSTOM_PID_FILE=""

# These variables tell us which action we will be performing.
START=0
STOP=0
QUERY=0

# Whether the process we're going to start backgrounds itself
BG_PROC=0
BG_PID_FILE=""

# Whether the execution should be quiet.
QUIET=0

# Original options
ORIG_OPTIONS=""

# Read the command line arguments and set variables accordingly.
# Does no error checking on the inputs.
while getopts "dnb:s:k:r:u:q:t:c:f:p:" option
do
    case "$option" in
        d ) DAEMONIZE=1;;
        n ) QUIET=1; ORIG_OPTIONS=${ORIG_OPTIONS}" -n";;
        b ) BG_PROC=1; BG_PID_FILE=$OPTARG; ORIG_OPTIONS=${ORIG_OPTIONS}" -b $OPTARG";;
        s ) START=1; TAG=$OPTARG; ORIG_OPTIONS=${ORIG_OPTIONS}" -s $OPTARG";;
        k ) STOP=1; TAG=$OPTARG;;
        r ) QUERY=1; TAG=$OPTARG;;
        u ) MIN_UPTIME=$OPTARG; ORIG_OPTIONS=${ORIG_OPTIONS}" -u $OPTARG";;
        q ) MAX_QUICK_FAILURES=$OPTARG; ORIG_OPTIONS=${ORIG_OPTIONS}" -q $OPTARG";;
        t ) MAX_TOTAL_FAILURES=$OPTARG; ORIG_OPTIONS=${ORIG_OPTIONS}" -t $OPTARG";;
        c ) CLEANUP_CMD=$OPTARG; ORIG_OPTIONS=${ORIG_OPTIONS}" -c $OPTARG";;
        f ) EXIT_CLEANUP_CMD=$OPTARG; ORIG_OPTIONS=${ORIG_OPTIONS}" -f $OPTARG";;
        p ) CUSTOM_PID_FILE=$OPTARG; ORIG_OPTIONS=${ORIG_OPTIONS}" -p $OPTARG";;
        * ) usage;;
    esac
done

# Everything after the options is the command we are running.
shift $(expr $OPTIND - 1)
CMD=$@

# The current number of failures.
QUICK_FAILURES=0
TOTAL_FAILURES=0

# Use this file to keep state across invocations of the script (i.e. allow
# us to query or kill a watchdog we start).
LABEL="watchdog-$TAG"
PIDDIR="/var/run/vmware/nsx-netopa"
PIDFILE="$PIDDIR/$LABEL.PID"
RESTART_FILE="/var/run/vmware/restart-$TAG"
ALWAYS_RESTART_FILE="/etc/vmware/always-restart-$TAG"

# Make sure the pid directory exists
mkdir -p -m 755 "${PIDDIR}"

if [ "${DAEMONIZE}" = "1" -a $START = 1 ] ; then
    daemonize
elif [ $START = 1 ] ; then
    start
elif [ $STOP = 1 ] ; then
    stop
elif [ $QUERY = 1 ] ; then
    query
else
    usage
fi
