# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import json
import os
import select
import sys
import threading
from queue import Queue, Full
from collections import deque

import sha.preconfig
from sha.cli import socket_utils
from sha.cli.cli_cmds import CLI_CMD_TABLE
from sha.cli.cli_utils import CLI_SERVICE_SOCKET, CLI_CONTROL_SOCKET, CLI_REQ_ID, CLI_REQ_CMD, CLI_REQ_ARGS, \
    CLI_REQ_HELP, CLI_SERVER_WORKERS, CommandResponse, CLI_SERVER_PIPE_MAGIC


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        return list(obj) if isinstance(obj, set) else (
            json.JSONEncoder.default(self, obj))


class CliWorker(threading.Thread):
    @classmethod
    def current_thread(cls):
        return threading.current_thread()

    def __init__(self, target=None, *args, **kwargs):
        threading.Thread.__init__(self, target=target, *args, **kwargs)
        self._killed = False

    def start(self):
        self.__run_backup = self.run
        self.run = self.__run
        threading.Thread.start(self)

    def __run(self):
        sys.settrace(self.global_trace)
        self.__run_backup()
        self.run = self.__run_backup

    def global_trace(self, frame, event, arg):
        if event == 'call':
            return self.local_trace
        else:
            return None

    def local_trace(self, frame, event, arg):
        if self._killed:
            if event == 'line':
                raise SystemExit()
        return self.local_trace

    def kill(self):
        self._killed = True


class CliServer:

    def __init__(self, server_name="sha.cli", srv_sock_file=CLI_SERVICE_SOCKET, ctl_sock_file=CLI_CONTROL_SOCKET,
                 cmd_table=CLI_CMD_TABLE):
        self._logger = sha.preconfig.get_sub_logger(__name__, 'cli-server')
        self._logger.info("CLI server initializing")
        self._server_name = server_name
        self._srv_sock = None
        self._srv_sock_file = srv_sock_file
        self._srv_ready = False
        self._srv_worker = None
        self._srv_pipes = None
        self._ctl_sock = None
        self._ctl_sock_file = ctl_sock_file
        self._ctl_ready = False
        self._ctl_worker = None
        self._ctl_pipes = None
        self._cmd_table = cmd_table
        self._request_queue = Queue(maxsize=CLI_SERVER_WORKERS)
        self._id_thread_table = {}
        self._id_conn_table = {}
        self._cancelled_id_queue = deque(maxlen=10)
        self._cancel_lock = threading.Lock()
        self._workers = set()

    def _accept_client(self, sock, sock_name, is_service, request_handler):
        conn = None
        pipe = None
        read_list = [sock]
        if is_service and self._srv_pipes:
            pipe = self._srv_pipes[0]
        elif not is_service and self._ctl_pipes:
            pipe = self._ctl_pipes[0]
        if pipe:
            read_list.append(pipe)
        try:
            while True:
                readable, _, _ = select.select(read_list, [], [])
                if is_service and not self._srv_ready:
                    break
                if not is_service and not self._ctl_ready:
                    break
                if sock in readable:
                    conn = socket_utils.accept_conn(sock)
                    request_handler(conn)
                    conn = None
                if pipe and pipe in readable:
                    magic = os.read(pipe, 1)
                    if magic == CLI_SERVER_PIPE_MAGIC:
                        break
        except Exception as e:
            self._logger.warning("Exception in accepting clients of socket {}: {}".format(sock_name, e))
        finally:
            self._logger.info("Closing socket {}".format(sock_name))
            conn_set = set(self._id_conn_table.values()) if is_service else set()
            conn_set.add(conn)
            for connection in conn_set:
                socket_utils.close_conn(connection)
            socket_utils.close_socket(sock)
            self._logger.info("Closed socket {}".format(sock_name))

    def _create_worker(self):
        thread = CliWorker(target=self._worker)
        self._workers.add(thread)
        thread.start()

    def _create_socket(self):
        try:
            self._srv_sock = socket_utils.create_and_bind_socket(self._srv_sock_file, 5)
        except Exception as e:
            self._logger.error("Failed to create socket for {}: {}".format(self._srv_sock_file, e))
            return False

        try:
            self._ctl_sock = socket_utils.create_and_bind_socket(self._ctl_sock_file, 5)
        except Exception as e:
            self._logger.error("Failed to create socket for {}: {}".format(self._ctl_sock_file, e))
            return False

        return True

    def _close_socket(self):
        socket_utils.close_socket(self._srv_sock)
        socket_utils.close_socket(self._ctl_sock)

    def _service(self):
        self._logger.info("Starting service thread for {}".format(self._server_name))

        self._srv_ready = True
        self._accept_client(self._srv_sock, str(self._srv_sock_file), True, self._dispatch_command)

        self._logger.info("Stopped service thread for {}".format(self._server_name))

    def _control(self):
        self._logger.info("Starting monitor thread for {}".format(self._server_name))

        self._ctl_ready = True
        self._accept_client(self._ctl_sock, str(self._ctl_sock_file), False, self._cancel_command)

        self._logger.info("Stopped monitor thread for {}".format(self._server_name))

    def _create_pipe(self):
        if sys.platform != 'win32':
            self._srv_pipes = os.pipe()
            self._ctl_pipes = os.pipe()

    def _notify_stop(self):
        if self._srv_pipes:
            os.write(self._srv_pipes[1], CLI_SERVER_PIPE_MAGIC)
        if self._ctl_pipes:
            os.write(self._ctl_pipes[1], CLI_SERVER_PIPE_MAGIC)

    def start(self):
        self._logger.info("Starting server {}".format(self._server_name))

        status = self._create_socket()
        if not status:
            self._close_socket()
            return

        for _ in range(CLI_SERVER_WORKERS):
            self._create_worker()

        self._create_pipe()

        self._srv_worker = threading.Thread(target=self._service)
        self._srv_worker.start()

        self._ctl_worker = threading.Thread(target=self._control)
        self._ctl_worker.start()

        self._logger.info("Started server {}".format(self._server_name))

    def stop(self):
        self._logger.info("Stopping server {}".format(self._server_name))

        for worker in self._workers:
            worker.kill()

        self._srv_ready = False
        self._ctl_ready = False
        # Notify service and monitor threads to stop polling socket
        self._notify_stop()
        self._close_socket()

        if self._srv_worker:
            self._srv_worker.join()
            self._srv_worker = None
        if self._ctl_worker:
            self._ctl_worker.join()
            self._ctl_worker = None

        self._logger.info("Stopped server {}".format(self._server_name))

    def _worker(self):
        current_thread = CliWorker.current_thread()
        while True:
            req = self._request_queue.get()
            req_id = req[CLI_REQ_ID]
            conn = self._id_conn_table[req_id]
            # There is a potential racing between worker thread and command cancel thread:
            #  - Worker thread checks if the current request has been cancelled
            #  - Command cancel thread adds request to be cancelled
            # Hence, a lock is added to make actions related to request cancelling atomic.
            with self._cancel_lock:
                self._id_thread_table[req_id] = current_thread
                is_cancelled = req_id in self._cancelled_id_queue
            if not is_cancelled:
                self._process_request(conn, req)
            socket_utils.close_conn(conn)
            self._id_thread_table.pop(req_id)
            self._id_conn_table.pop(req_id)

    def _dispatch_command(self, conn):
        msg, err_msg = socket_utils.receive_msg(conn)
        if err_msg:
            self._logger.error(err_msg)
            return self._send_response(conn, False, None, err_msg)

        try:
            req = json.loads(msg)
        except ValueError as e:
            err_msg = "Received malformed request {}: {}".format(msg, e)
            self._logger.error(err_msg)
            return self._send_response(conn, False, None, err_msg)

        self._logger.info("Received request {}".format(req))

        err_msg = ""
        if CLI_REQ_ID not in req:
            err_msg = "Missing {} in request".format(CLI_REQ_ID)
        elif CLI_REQ_CMD not in req:
            err_msg = "Missing {} in request".format(CLI_REQ_CMD)
        if err_msg:
            self._logger.error(err_msg)
            return self._send_response(conn, False, req, err_msg)

        req_id = req[CLI_REQ_ID]
        self._id_conn_table[req_id] = conn
        try:
            self._request_queue.put(req, False)
        except Full:
            err_msg = "CLI server has task queue full"
            self._logger.error(err_msg)
            return self._send_response(conn, False, req, err_msg)

    def _process_request(self, conn, req):
        cmd_name = req[CLI_REQ_CMD]
        cmd_entry = self._cmd_table.get(cmd_name, None)
        if cmd_entry is None:
            err_msg = "Command {} is not supported".format(cmd_name)
            self._logger.error(err_msg)
            return self._send_response(conn, False, req, err_msg)

        args = req.get(CLI_REQ_ARGS, {})
        cmd = cmd_entry(args)
        is_help = args.get(CLI_REQ_HELP, False)
        try:
            if is_help:
                result = cmd.help()
            else:
                result = cmd.execute()
        except Exception as e:
            err_msg = "Exception occurred: {}".format(e)
            self._logger.error(err_msg)
            return self._send_response(conn, False, req, err_msg)

        return self._send_response(conn, True, req, result)

    def _send_response(self, conn, status, req, resp):
        if not status:
            resp = CommandResponse(None, -1, resp)
        try:
            status, err_msg = socket_utils.send_msg(
                conn, json.dumps(resp.payload(), cls=SetEncoder))
            if not status:
                self._logger.error(err_msg)
        except Exception as e:
            self._logger.error("Failed to send response for request {}: {}".format(
                "" if req is None else req, e))

    def _cancel_command(self, conn):
        msg, err_msg = socket_utils.receive_msg(conn)
        if err_msg:
            socket_utils.close_conn(conn)
            return

        try:
            req = json.loads(msg)
        except ValueError as e:
            self._logger.error("Received malformed request {}: {}".format(msg, e))
            return

        if CLI_REQ_ID not in req:
            self._logger.error("Missing {} in request".format(CLI_REQ_ID))
            return

        req_id = req[CLI_REQ_ID]
        self._logger.info("Received cancel request for {}".format(req_id))
        with self._cancel_lock:
            thread = self._id_thread_table.pop(req_id, None)
            if not thread:
                # The command request is not scheduled to a worker yet.
                # Such request will be cleaned up when it is scheduled.
                if req_id not in self._cancelled_id_queue:
                    self._cancelled_id_queue.append(req_id)
                return
        thread.kill()
        self._workers.remove(thread)
        self._create_worker()
        client_conn = self._id_conn_table.pop(req_id, None)
        socket_utils.close_conn(client_conn)
        socket_utils.close_conn(conn)
        self._logger.info("Finished cancel request for {}".format(req_id))
